/*
     Copyright (C) 2004, 2020 Hubert Vansteenhuyse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "subnet.h"


#define CHAR_HLINE '-'		/*fix: sometimes ACS_... are not supported */
#define CHAR_VLINE '|'
#define CHAR_ULCORNER '+'
#define CHAR_URCORNER '+'
#define CHAR_LLCORNER '+'
#define CHAR_LRCORNER '+'

int sw_exit = 0;


#ifdef __LINUX__
WINDOW *wnd = NULL;
#endif

#ifdef __WINDOWS__
int sw_invert_screen = 0;
HWND hwndMain;
WNDCLASSEX wndclass;
char *szMainWndClass = "SubnetCalculator";
int __cursor_y = 0, __cursor_x = 0;

LRESULT CALLBACK
MainWndProc(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam){
    switch (nMsg) {
    case WM_MOUSEWHEEL:
        if(GET_WHEEL_DELTA_WPARAM(wParam) > 0)    		win_kbd = (char) PGU;
        else    		win_kbd = (char) PGD;
        break;

    case WM_DESTROY:
        sw_exit = 1;
        win_kbd = '\t';		//try to leave the input routine...
        PostQuitMessage(0);
        return 0;
        break;

    case WM_PAINT:
        updateScreen();
        break;

    case WM_CHAR:
        win_kbd = (char) wParam;
        return 0;
        break;

    case WM_KEYDOWN:
        switch (wParam) {
        case VK_PRIOR:
            win_kbd = (char) PGU;
            break;
        case VK_NEXT:
            win_kbd = (char) PGD;
            break;
        case VK_LEFT:
            win_kbd = (char) KEYL;
            break;
        case VK_RIGHT:
            win_kbd = (char) KEYR;
            break;
        case VK_UP:
            win_kbd = (char) KEYU;
            break;
        case VK_DOWN:
            win_kbd = (char) KEYD;
            break;
        case VK_BACK:
            win_kbd = (char) KEYBACK;
            break;
        }
        return 0;
        break;
    }

    return DefWindowProc(hwnd, nMsg, wParam, lParam);
}

#endif


FILE *fout;
char *columns[] = { "#", "NET", "BCAST", "RANGE" };
int x_colpos[4];

char foundres[128];
char buf[1024 * 2], ipbuf[32];

int THESCREEN = 0,		//the current screen
    THETAB = 0,			//selector for cidr, mask, subnet input
    taborder = 0,		//input selector (CIDR, subnets, mask)
    Subnets = 0;

int x_mask = 0, x_hosts = 0, box_canceled = 0, dlgbox_canceled = 0, x_wild = 0;

int ycenter = 0, xcenter = 0, helpbegin = 0;

long netstartcounter = 1;

char *about[] = {
    // l = left, r = right, c center, - line
    "cABOUT SUBNET CALCULATOR",
#ifdef __WINDOWS__
    "cWINDOWS NATIVE GDI PORT",
#else
    "cUNIX/LINUX TERMINAL PORT",
#endif
    "cTool to calculate TCP/IP version 4 subnets and supernetworks.",
    "-",
    "lCopyright (C) 2004, 2020  Vansteenhuyse Hubert",
    "lThis program is free software; you can redistribute it and/or modify",
    "lit under the terms of the GNU General Public License as published by",
    "lthe Free Software Foundation; either version 2 of the License, or",
    "l(at your option) any later version.",
    "lThis program is distributed in the hope that it will be useful,",
    "lbut WITHOUT ANY WARRANTY; without even the implied warranty of",
    "lMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
    "lGNU General Public License for more details.",
    "lYou should have received a copy of the GNU General Public License",
    "lalong with this program; if not, write to the Free Software",
    "lFoundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA",
    "-",
    "cREQUESTS/BUGS: hubert.vansteenhuyse@freecode.be",
    "cHOMEPAGE: http://www.freecode.be/subnet.htm",
    "c[PRESS ANY KEY]",
    "c",
    NULL
};


char *helper[] = {
    // l = left, r = right, c center, - line
    "cHELP",
    "-",
    "lMAIN SCREEN:",
    "l    Press a NUMBER key to change the value at the cursor.",
    "l    TAB/LEFT/RIGHT............moves the cursor between the header parameters",
    "l                              CIDR, NETMASK, WILD, #NETWORKS and #HOSTS",
    "l    UP/DOWN/PAGEUP/PAGEDOWN...scrolls the subnet list",
    "l    SPACE.....................redraws the screen (useful in terminals)",
    "l    a.........................about screen",
    "l    c.........................command line options",
    "l    d.........................the dictionary",
    "l    f.........................finds the subnet of an ip address",
    "l    j.........................jumps to a subnet",
    "l    h.........................this help screen",
    "l    q.........................quits the program",
    "l    w.........................writes the listing to a file",
    "l",
    "-",
    "lDIALOG BOXES:",
    "l    - Type the value using any NUMBER keys and the '.' key.",
    "l    - Key ENTER accepts the value. Keys TAB/ESCAPE cancels the value.",
    "l    - Keys DELETE/BACKSPACE removes the character before the cursor.",
    "l",
    "-",
    "c[PRESS ANY KEY]",
    "c",
    NULL
};

char *dictionary[] = {
    // l = left, r = right, c center, - line
    "cDICTIONARY",
    "-",
    "lSUBNETTING:",
    "lBorrowing some bits of the host part from an ip address to create more networks",
    "lis subnetting. As a result a large network will be separated in smaller subnets.",
    "lAnother method is transparent bridging which operates on OSI LAYER 2 while",
    "lsubnetting operates on LAYER 3.",
    "l",
    "lSUPERNETWORKING",
    "lDefines access to one or more subnets or hosts by using 'wildcards', represented",
    "las inverted subnet mask. Binary speaking a 0 means compare and a 1 means ignore.",
    "lE.g.: 00001111 means use the first 4 bits for compare and don't use the next 4.",
    "l",
    "lCIDR (Classless Inter Domain Routing).",
    "lThis notation is a way to represent an ip address and it's subnet mask.",
    "lConventional: 127.0.0.1 255.255.0.0 CIDR: 127.0.0.1 / 16",
    "l",
    "lMAGIC NUMBER:",
    "lThis is 256 minus the last nonzero dot separated digit group in the mask,",
    "lwhich can be used to calculate the subnet chunks. For instance the magic number",
    "lfor 255.255.255.254 is 256-254=2, for 255.224.0.0 is 256-224=32",
    "l",
    "c[PRESS ANY KEY]",
    "c",
    NULL
};

#ifdef __WINDOWS__		/* Emulate the used ncurse routines for Windows only... */
int move(int y, int x){
    HDC hdc;
    int _y, _x;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;

    if(sw_exit) return 1;

    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
    GetTextMetrics(hdc, &tm);

    _y = (tm.tmHeight + tm.tmExternalLeading) * (y + 1);
    _x = tm.tmAveCharWidth * x;
    SetCaretPos(_x, _y);
    __cursor_y = y;
    __cursor_x = x;

    SelectObject(hdc, oldobj);
    ReleaseDC(hwndMain, hdc);
    return 0;			//return ok
}

/*
	emulation of the UNIX curses 'mvaddstr' function.
    The string is printed to pos xy & the cursor is moved to it's tail.
*/
int mvaddstr(int y, int x, const char *str){
    HDC hdc;
    int len;
    HGDIOBJ oldobj;
    int _y, _x;
    TEXTMETRIC tm;

    if (sw_exit) return 1;

    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));

    SetBkMode(hdc, OPAQUE);
    if (sw_invert_screen) {
        SetTextColor(hdc, RGB(255, 255, 255));
        SetBkColor(hdc, RGB(0, 0, 0));
    } else {
        SetTextColor(hdc, RGB(0, 0, 0));
        SetBkColor(hdc, RGB(255, 255, 255));
    }

    len = strlen(str);
    GetTextMetrics(hdc, &tm);
    _y = (tm.tmHeight + tm.tmExternalLeading) * y;
    _x = tm.tmAveCharWidth * x;
    TextOut(hdc, _x, _y, str, len);
    move(y, x + len);

    SelectObject(hdc, oldobj);
    ReleaseDC(hwndMain, hdc);
    return 0;			//return ok
}

/*
    emulation of the UNIX curses 'addstr' function.
    The string is printed & the cursor is moved to it's tail.
*/
int addstr(const char *str){
    HDC hdc;
    int len = strlen(str);
    HGDIOBJ oldobj;
    int _y, _x;
    TEXTMETRIC tm;

    if (sw_exit) return 1;
    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));

    SetBkMode(hdc, OPAQUE);
    if (sw_invert_screen) {
        SetTextColor(hdc, RGB(255, 255, 255));
        SetBkColor(hdc, RGB(0, 0, 0));
    } else {
        SetTextColor(hdc, RGB(0, 0, 0));
        SetBkColor(hdc, RGB(255, 255, 255));
    }

    GetTextMetrics(hdc, &tm);
    _y = (tm.tmHeight + tm.tmExternalLeading) * __cursor_y;
    _x = tm.tmAveCharWidth * __cursor_x;

    TextOut(hdc, _x, _y, str, len);

    move(__cursor_y, __cursor_x + len);
    SelectObject(hdc, oldobj);
    ReleaseDC(hwndMain, hdc);
    return 0;			//return ok
}

/*
    emulation of the UNIX curses 'addch' function.
     The string is printed to pos xy & the cursor is moved to it's tail.
*/
int addch(const char c){
    HDC hdc;
    int len = 1;
    HGDIOBJ oldobj;
    int _y, _x;
    TEXTMETRIC tm;

    if (sw_exit) return 1;
    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));

    SetBkMode(hdc, OPAQUE);
    if (sw_invert_screen) {
        SetTextColor(hdc, RGB(255, 255, 255));
        SetBkColor(hdc, RGB(0, 0, 0));
    } else {
        SetTextColor(hdc, RGB(0, 0, 0));
        SetBkColor(hdc, RGB(255, 255, 255));
    }

    GetTextMetrics(hdc, &tm);
    _y = (tm.tmHeight + tm.tmExternalLeading) * __cursor_y;
    _x = tm.tmAveCharWidth * __cursor_x;
    TextOut(hdc, _x, _y, &c, len);
    move(__cursor_y, __cursor_x + 1);

    SelectObject(hdc, oldobj);
    ReleaseDC(hwndMain, hdc);
    return 0;			//return ok
}

/*
Clears the screen
*/
void clear(void){
    HDC hdc;
    RECT rc;
    if(sw_exit) return;
    hdc = GetDC(hwndMain);
    GetClientRect(hwndMain, &rc);
    Rectangle(hdc, rc.left - 10, rc.top - 10, rc.right + 10, rc.bottom + 10);
    move(0, 0);
    ReleaseDC(hwndMain, hdc);
}

/*
Clears the line from right side of the cursor
*/
void clrtoeol(void){
    HDC hdc;
    RECT rc;
    HGDIOBJ oldobj, oldPen;
    TEXTMETRIC tm;

    if (sw_exit) return;

    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
    GetTextMetrics(hdc, &tm);

    GetClientRect(hwndMain, &rc);
    rc.left = tm.tmAveCharWidth * get_x();
    rc.top = (tm.tmHeight + tm.tmExternalLeading) * get_y();
    rc.bottom = rc.top + tm.tmHeight + tm.tmExternalLeading;
    oldPen = SelectObject(hdc, GetStockObject(WHITE_PEN));
    Rectangle(hdc, rc.left, rc.top, rc.right, rc.bottom);
    SelectObject(hdc, oldPen);

    SelectObject(hdc, oldobj);
    ReleaseDC(hwndMain, hdc);
}

#endif /*__WINDOWS__*/


/* Reverses the screen colors. sw_switch: 0 = normal, 1 = reversed */
void reverse(int sw_switch){
#ifdef __LINUX__
    attrset((sw_switch == 0 ? WA_NORMAL : WA_REVERSE));
#endif

#ifdef __WINDOWS__
    sw_invert_screen = sw_switch;
#endif
}

void cursor_on(int sw_turn){
#ifdef __WINDOWS__
    HDC hdc = 0;
    TEXTMETRIC tm;
    HGDIOBJ oldobj;
    if (sw_exit)
        return;
    if (sw_turn) {
        hdc = GetDC(hwndMain);
        if(!hdc) return;

        oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
        GetTextMetrics(hdc, &tm);
        SelectObject(hdc, oldobj);
        CreateCaret(hwndMain, (HBITMAP) NULL, tm.tmAveCharWidth, tm.tmHeight / 6);
        ShowCaret(hwndMain);
    } else {
        HideCaret(hwndMain);
    }
    ReleaseDC(hwndMain, hdc);

#endif /*__WINDOWS__*/
}

/* ** screen area ********************************************************* */
/* returns the x position of the cursor */
int get_x(void){
    int retval = 0;
#ifdef __LINUX__
    retval = wnd->_curx;
#endif

#ifdef __WINDOWS__
    retval = __cursor_x;
#endif	   /*__WINDOWS__*/
    return retval;
}

/* returns the y position of the cursor */
int get_y(void){
    int retval = 0;
#ifdef __LINUX__
    retval = wnd->_cury;
#endif

#ifdef __WINDOWS__
    retval = __cursor_y;
#endif	   /*__WINDOWS__*/
    return retval;
}

/* Returns the width of the screen area*/
int get_maxx(void){
    int retval = 0;
#ifdef __LINUX__
    retval = wnd->_maxx;
#endif

#ifdef __WINDOWS__
    HDC hdc;
    TEXTMETRIC tm;
    RECT rc;
    if (sw_exit)
        return 1;
    HGDIOBJ oldobj;
    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
    GetTextMetrics(hdc, &tm);
    GetClientRect(hwndMain, &rc);
    SelectObject(hdc, oldobj);
    retval = (int) ((rc.right - rc.left) / tm.tmAveCharWidth);
    ReleaseDC(hwndMain, hdc);
#endif	   /*__WINDOWS__*/
    return retval;
}

/* Returns the height of the screen area*/
int get_maxy(void){
    int retval = 0;
#ifdef __LINUX__
    retval = wnd->_maxy;
#endif

#ifdef __WINDOWS__
    HDC hdc;
    TEXTMETRIC tm;
    RECT rc;
    HGDIOBJ oldobj;
    unsigned int my;
    if (sw_exit)	return 1;
    hdc = GetDC(hwndMain);
    oldobj = SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
    GetTextMetrics(hdc, &tm);
    GetClientRect(hwndMain, &rc);
    SelectObject(hdc, oldobj);
    my = (rc.bottom - rc.top) / (tm.tmHeight + tm.tmExternalLeading);
    retval =
        my - ((my * (tm.tmHeight + tm.tmExternalLeading)) >
              (rc.bottom - rc.top - ((tm.tmHeight * 3) / 4)) ? 1 : 0);
    ReleaseDC(hwndMain, hdc);
#endif	   /*__WINDOWS__*/
    return retval;
}

/* ************************************************************************************************************************
**************************************************************************************************************************
*************************************************************************************************************************/
void reInitScreen(){
#ifdef __LINUX__
    endwin();
    wnd = initscr();
#endif
}

void SIGNAL_RESIZE(int n){
#ifdef __LINUX__
    reInitScreen();
    updateScreen();
    refresh();
#endif
}

int center(int y, char *str){
    int x = (get_maxx() - strlen(str)) / 2;
    mvaddstr(y, x, str);
    return x;
}
int left(int y, char *str){
    mvaddstr(y, 0, str);
    return 0;
}
int right(int y, char *str){
    int x = 0;
    x = get_maxx() - strlen(str);
#ifdef __WINDOWS__
    x = x - 1;
#else
    x = x + 1;
#endif
    mvaddstr(y, get_maxx(), " ");
    mvaddstr(y, x, str);
    return x;
}

char *format(char *fmt, ...){
    va_list args;
    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);
    return buf;
}

void drawhorline(int y){
    int c;
    move(y, 0);
    for (c = 0; c <= get_maxx(); c++)
        addch(CHAR_HLINE);
    move(y, 0);
}

void drawverline(int y, int x){
    int _x = x, _y = y;
    while (y <= get_maxy()) {
        move(y, x);
        addch(CHAR_VLINE);
        y++;
    }
    move(_y, _x);
}

void drawbox(int y, int x, int height, int width){
    int v, h;

    for (v = y; v < y + height; v++) {
        move(v, x);
        for (h = 0; h <= width; h++)
            addch(' ');
    }

    move(y, x);
    addch(CHAR_ULCORNER);
    for (h = 0; h <= width - 2; h++)
        addch(CHAR_HLINE);
    addch(CHAR_URCORNER);
    move(y + 2, x);
    for (h = 0; h <= width; h++)
        addch(CHAR_HLINE);
    move(y + height, x);
    addch(CHAR_LLCORNER);
    for (h = 0; h <= width - 2; h++)
        addch(CHAR_HLINE);
    addch(CHAR_LRCORNER);
    for (v = y + 1; v < y + height; v++) {
        move(v, x);
        addch(CHAR_VLINE);
    }
    for (v = y + 1; v < y + height; v++) {
        move(v, x + width);
        addch(CHAR_VLINE);
    }
    move(y + 1, x + 2);
}

//inputbox for ip addresses **************************************************************************************
static char ipbox_title[64];
static char ipbox_buf[64];
#define WIDTH_IPBOX 48
char *ipbox(char *title, char firstdigit){
    char in;
    ycenter = (get_maxy() - 6) / 2;
    xcenter = (get_maxx() - WIDTH_IPBOX) / 2;

    if (title != NULL) {
        box_canceled = 0;
        sprintf(ipbox_title, " %s", title);
        memset(ipbox_buf, 0, 64);
        sprintf(ipbox_buf, "%c", firstdigit);
    }

    reverse(1);
    drawbox(ycenter, xcenter, 7, WIDTH_IPBOX);
    addstr(ipbox_title);
    mvaddstr(ycenter + 6, xcenter + 2, "keys: enter esc backspace");

    if (title != NULL) {
        for (;;) {
            mvaddstr(ycenter + 3, xcenter + 9, format("%s", getiptobin(ipbox_buf) ));
            mvaddstr(ycenter + 4, xcenter + 2, format("VALUE: %s", ipbox_buf));
            in = getinput();
            if (in == ' ') {
                updateScreen();
                continue;
            }
            if (in == '\t' || in == 27) {
                box_canceled = 1;
                break;
            }
            if (in == 13 || in == 10) {
                break;
            }
            if (in == KEYBACK && strlen(ipbox_buf) > 0) {
                ipbox_buf[strlen(ipbox_buf) - 1] = 0;
                mvaddstr(ycenter + 4, xcenter + 2, format("VALUE: %s ", ipbox_buf));
            }
            if ((strlen(ipbox_buf) < 15) && (in == '.' || ISDIGIT(in))) {
                sprintf(ipbox_buf + strlen(ipbox_buf), "%c", in);
            }

        }
    }


    if (title != NULL) memset(ipbox_title, 0, 64);
    reverse(0);
    return ipbox_buf;
}

//inputbox for digits **************************************************************************************
static char intbox_title[64];
static char intbox_msg[128];
static char intbox_buf[8];
char *intbox(char *title, char *msg, char firstdigit){
    int wdt;
    char in;

    reverse(1);
    if (title != NULL) {
        box_canceled = 0;
        sprintf(intbox_title, "%s", title);
        sprintf(intbox_msg, "%s", msg);
        memset(intbox_buf, 0, 8);
        sprintf(intbox_buf, "%c", firstdigit);
    }

    wdt = strlen(intbox_title) + 4;
    if (strlen(intbox_msg) > wdt)
        wdt = strlen(intbox_msg) + 4;

    ycenter = (get_maxy() - 7) / 2;
    xcenter = (get_maxx() - wdt) / 2;

    drawbox(ycenter, xcenter, 7, wdt);
    addstr(intbox_title);
    mvaddstr(ycenter + 3, xcenter + 2, intbox_msg);
    mvaddstr(ycenter + 6, xcenter + 2, "keys: enter esc backspace");
    mvaddstr(ycenter + 4, xcenter + 2, format("VALUE: %s", intbox_buf));

    if (title != NULL) {
        for (;;) {
            in = getinput();
            if (in == ' ') {
                updateScreen();
                continue;
            }
            if (in == '\t' || in == 27) {
                box_canceled = 1;
                break;
            }
            if (in == 13 || in == 10) {
                break;
            }
            if (in == KEYBACK && strlen(intbox_buf) > 0) {
                intbox_buf[strlen(intbox_buf) - 1] = 0;
                mvaddstr(ycenter + 4, xcenter + 2,
                         format("VALUE: %s ", intbox_buf));
                mvaddstr(ycenter + 4, xcenter + 2,
                         format("VALUE: %s", intbox_buf));
            }
            if ((strlen(intbox_buf) < 8) && ISDIGIT(in)) {
                sprintf(intbox_buf + strlen(intbox_buf), "%c", in);
                mvaddstr(ycenter + 4, xcenter + 2,
                         format("VALUE: %s", intbox_buf));
            }
        }
    }

    if (title != NULL)
        memset(intbox_title, 0, 64);
    reverse(0);
    return intbox_buf;
}


//Messagebox **************************************************************************************
static char mbox_title[64];
static char mbox_msg[128];
void messagebox(char *title, char *msg){
    int wdt;
    reverse(1);
    if (title) {
        sprintf(mbox_title, "%s", title);
        sprintf(mbox_msg, "%s", msg);
    }
    wdt = strlen(mbox_title) + 4;
    if (strlen(mbox_msg) > wdt)
        wdt = strlen(mbox_msg) + 4;
    ycenter = (get_maxy() - 5) / 2;
    xcenter = (get_maxx() - wdt) / 2;

    drawbox(ycenter, xcenter, 5, wdt);
    addstr(mbox_title);
    mvaddstr(ycenter + 3, xcenter + 2, mbox_msg);
    mvaddstr(ycenter + 4, xcenter + 6, "Press any key...");

    if (title) {
        getinput();
        memset(mbox_title, 0, 64);
    }
    reverse(0);
}

//dialoogbox voor alle karakters **************************************************************************************
static char dlgbox_title[64];
static char dlgbox_buf[128];
void dlgbox(char *title){
    int wdt, start, n;
    char in;

    reverse(1);
    if (title != NULL) {
        dlgbox_canceled = 0;
        sprintf(dlgbox_title, "%s", title);
    }

    memset(dlgbox_buf, 0, 128);
    wdt = get_maxx();
    if (wdt > 128 / 2)
        wdt = 128 / 2;

    ycenter = (get_maxy() - 5) / 2;
    xcenter = (get_maxx() - wdt) / 2;

    drawbox(ycenter, xcenter, 5, wdt);
    addstr(dlgbox_title);
    mvaddstr(ycenter + 4, xcenter + 2, "keys: enter esc backspace");
    if (strlen(dlgbox_buf) > (wdt - 4))
        start = strlen(dlgbox_buf) - wdt + 4;
    else
        start = 0;
    mvaddstr(ycenter + 3, xcenter + 2, dlgbox_buf + start);

    if (title != NULL) {
        for (;;) {
            in = getinput();
            if (in == ' ') {
                updateScreen();	/*continue; */
            }
            if (in == '\t' || in == 27) {
                dlgbox_canceled = 1;
                break;
            }
            if (in == 13 || in == 10) {
                break;
            }
            if(in == KEYBACK) {
                start = 0;
                n = strlen(dlgbox_buf);
                if(n > 0) {
                    dlgbox_buf[n] = 0;
                    n--;
                    dlgbox_buf[n] = 0;
                    if(n > (wdt - 4)) start = n - wdt + 4;
                }
                mvaddstr(ycenter + 3, xcenter + 2, format("%s ", dlgbox_buf + start));
                mvaddstr(ycenter + 3, xcenter + 2, format("%s", dlgbox_buf + start));
                continue;
            }

            if(strlen(dlgbox_buf) < 128 && in >= ' ' && in <= '{') {
                n = strlen(dlgbox_buf);
                dlgbox_buf[n] = in;
                n++;
                if (n > (wdt - 4))
                    start = n - wdt + 4;
                else
                    start = 0;
                mvaddstr(ycenter + 3, xcenter + 2, dlgbox_buf + start);
            }
        }
    }

    if (title != NULL)
        memset(dlgbox_title, 0, 64);
    reverse(0);
}

/*************************************************************************************************************************
**************************************************************************************************************************
*************************************************************************************************************************/
void exitScreen(void){
#ifdef __LINUX__
    endwin();
#endif
}

int initScreen(void){
#ifdef __LINUX__
    if (NULL != (wnd = initscr())) {
        keypad(stdscr, TRUE);
        noecho();
        signal(SIGWINCH, SIGNAL_RESIZE);
        memset(&buf, 0, 1024);
        return 1;

    }
#endif				/* __LINUX__ */

#ifdef __WINDOWS__
    memset(&wndclass, 0, sizeof(WNDCLASSEX));
    wndclass.lpszClassName = szMainWndClass;
    wndclass.cbSize = sizeof(WNDCLASSEX);
    wndclass.style = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc = MainWndProc;
    wndclass.hbrBackground = GetStockObject(WHITE_BRUSH);
    wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
    RegisterClassEx(&wndclass);

    hwndMain =
        CreateWindow(szMainWndClass, szMainWndClass, WS_OVERLAPPEDWINDOW,
                     CW_USEDEFAULT, CW_USEDEFAULT, 800, 600,/*CW_USEDEFAULT, CW_USEDEFAULT,*/
                     NULL, NULL, NULL, NULL);
    if (hwndMain) {
        ShowWindow(hwndMain, SW_SHOWNORMAL);
        UpdateWindow(hwndMain);
        return 1;
    }
#endif	/*__WINDOWS__*/

    return 0;
}

void input(void){
    int in, val;
    long lval;
    while (sw_exit == 0) {
        in = getinput();
        if (!in)
            continue;

        if (THESCREEN == 0 && ISDIGIT(in)) {
            switch (THETAB) {
            case 0:		//cidr
                if (isvalidip(ipbox("IP ADRESS", in))) {
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    ip = iptolong(ipbox_buf);
                    netstartcounter = 1;
                    switch ((int) getipclass(ip)) {
                    case (int) 'A':
                        mask = iptolong("255.0.0.0");
                        wild = mask_inverse(mask);
                        break;

                    case (int) 'B':
                        mask = iptolong("255.255.0.0");
                        wild = mask_inverse(mask);
                        break;

                    case (int) 'C':
                        mask = iptolong("255.255.255.0");
                        wild = mask_inverse(mask);
                        break;
                    }
                } else {
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    updateScreen();
                    messagebox(format("INVALID IP \"%s\"", ipbox_buf),
                               "Use digits 0-255, format x.y.w.z");
                }
                updateScreen();
                break;

            case 1:		//cidr mask
                switch ((int) getipclass(ip)) {
                case (int) 'A':
                    val =
                        atoi(intbox
                             ("SUBNETMASK (8, 10 to 30)", "CIDR netmask input", in));
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    if (val == 8 || (val >= 10 && val <= 30)) {
                        mask = cidrtomask(val);
                        wild = mask_inverse(mask);
                        netstartcounter = 1;
                    } else {
                        updateScreen();
                        messagebox(format
                                   ("INVALID MASK \"%s\"", intbox_buf),
                                   "The IP is class A: Valid masks are 8 or from 10 to 30");
                    }
                    break;

                case (int) 'B':
                    val =
                        atoi(intbox
                             ("SUBNETMASK",
                              "Valid masks are 16 or from 18 to 30", in));
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    if (val == 16 || (val >= 18 && val <= 30)) {
                        mask = cidrtomask(val);
                        wild = mask_inverse(mask);
                        netstartcounter = 1;
                    } else {
                        updateScreen();
                        messagebox(format
                                   ("INVALID MASK \"%s\"", intbox_buf),
                                   "The IP is class B: Valid masks are 16 or from 18 to 30");
                    }
                    break;

                case (int) 'C':
                    val =
                        atoi(intbox
                             ("SUBNETMASK",
                              "Valid masks are 24 or from 26 to 30", in));
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    if (val == 24 || (val >= 26 && val <= 30)) {
                        mask = cidrtomask(val);
                        wild = mask_inverse(mask);
                        netstartcounter = 1;
                    } else {
                        updateScreen();
                        messagebox(format
                                   ("INVALID MASK \"%s\"", intbox_buf),
                                   "The IP is class C: Valid masks are 24 or from 16 to 30");
                    }
                    break;
                default:
                    val =
                        atoi(intbox
                             ("New subnetmask",
                              "TODO: identify valid masks CLASS D & F",
                              in));
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    mask = cidrtomask(val);
                    wild = mask_inverse(mask);

                }
                updateScreen();
                break;

            case 2:		//mask format 255.255.0.0
                if (isvalidmask(iptolong(ipbox("SUBNETMASK (255,254,252,248,224,192,128)", in)))) {
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    mask = iptolong(ipbox_buf);
                    wild = mask_inverse(mask);
                    netstartcounter = 1;
                } else {
                    if (box_canceled) {
                        updateScreen();
                        continue;
                    }
                    updateScreen();
                    messagebox(format
                               ("INVALID SUBNETMASK \"%s\"", ipbox_buf),
                               "Invalid for network class or incorrect continuity of bits");
                }
                updateScreen();
                break;
            case 3:		//#wildcards / inversed netmask
                ipbox("Wildcards Filtering (3,7,15,31,63,127,255)", in);
                if(box_canceled) {
                    updateScreen();
                    continue;
                }
                if(strlen(ipbox_buf) == 0)
                    wild = mask_inverse(mask);
                else {
                    lval = iptolong(ipbox_buf);
                    if(isvalidmask_inverse(lval)) {
                        wild = lval;
                        mask = mask_inverse(wild);
                    } else {
                        updateScreen();
                        messagebox(format("INVALID WILDCARD \"%s\"", ipbox_buf), "Invalid for network class or incorrect continuity of bits");
                    }
                }
                updateScreen();
                break;
            case 4:		//#subnets
                lval =
                    atoi(intbox
                         ("# NETWORKS",
                          "The nearest amount of possible subnets will be used",
                          in));
                if (!box_canceled) {
                    useamount_net(lval);
                    netstartcounter = 1;
                }
                updateScreen();
                break;
            case 5:		//#hosts
                lval =
                    atoi(intbox
                         ("# HOSTS",
                          "The nearest amount of possible hosts will be used",
                          in));
                if (!box_canceled) {
                    useamount_host(lval);
                    netstartcounter = 1;
                }
                updateScreen();
                break;
            }
        }


        in = LCASE(in);

        if (in == 'a') {
            THESCREEN = 3;
            updateScreen();
            getinput();
            THESCREEN = 0;
            updateScreen();
            continue;
        }
        if (in == 'q') {
            sw_exit = 1;
            continue;
        }
        if (in == ' ') {
            updateScreen();
            continue;
        }
        if (in == 'h') {
            THESCREEN = 1;
            updateScreen();
            getinput();
            THESCREEN = 0;
            updateScreen();
            continue;
        }
        if (in == 'd') {
            THESCREEN = 2;
            updateScreen();
            getinput();
            THESCREEN = 0;
            updateScreen();
            continue;
        }
        if (in == 'c') {
            THESCREEN = 4;
            updateScreen();
            getinput();
            THESCREEN = 0;
            updateScreen();
            continue;
        }
        if (in == '\t' || in == KEYR) {
            THETAB++;
            if (THETAB > 5) {
                THETAB = 0;
            }
            updateScreen();
            continue;
        }
        if (in == KEYL) {
            THETAB--;
            if (THETAB < 0) {
                THETAB = 5;
            }
            updateScreen();
            continue;
        }

        if (in == 'w') {	//write to file
            dlgbox("Write to file. Type the file path");
            if (dlgbox_canceled) {
                updateScreen();
                continue;
            }
            fout = fopen(dlgbox_buf, "w");
            updateScreen();
            if (fout) {
                writefile(fout);
                updateScreen();
                fclose(fout);
            } else {
                messagebox("FILE ERROR",
                           format("File \"%s\" write failed. Error %d.",
                                  dlgbox_buf, errno));
                updateScreen();
            }
        }

        if (in == 'j') {	//jump
            lval =
                atoi(intbox
                     ("JUMP", "Type the network you want to see", 0));
            if (!box_canceled) {
                if (lval < 1)
                    lval = 1;
                if (lval > getnets())
                    lval = getnets();
                netstartcounter = lval;
            }
            updateScreen();
            continue;
        }

        if (in == 'f') {	//find the network of the ip
            if (!isvalidip(ipbox("Find IP address", 0))) {
                if (box_canceled) {
                    updateScreen();
                    continue;
                }
                updateScreen();
                messagebox(format("INVALID IP \"%s\"", ipbox_buf),
                           "Use digits(0-255) with format x.y.w.z");
            } else {
                if (!box_canceled) {
                    netstartcounter = findnet(iptolong(ipbox_buf));
                    if (netstartcounter == -1) {
                        messagebox("SUBNET NOT FOUND",
                                   format
                                   ("Address %s belongs to a different network.",
                                    ipbox_buf));
                        netstartcounter = 1;
                    } else {
                        if (netstartcounter < 1)
                            netstartcounter = 1;
                        sprintf(foundres, "[IP %s IS IN SUBNET %d]",
                                ipbox_buf, (unsigned) netstartcounter);
                    }
                }
            }
            updateScreen();
            continue;
        }

        if (in == KEYD) {
            if (netstartcounter + 1 > getnets())
                netstartcounter = getnets();
            else
                netstartcounter++;
            updateScreen();
            continue;
        }
        if (in == KEYU) {
            if (netstartcounter == 1)
                netstartcounter = 1;
            else
                netstartcounter--;
            updateScreen();
            continue;
        }

        if (in == PGD) {
            if (get_maxy() - 7 < 0)
                continue;
            if (netstartcounter + get_maxy() - 7 > getnets()) {
                while (netstartcounter > getnets()) {
                    netstartcounter++;
                }
            } else
                netstartcounter += get_maxy() - 7;
            updateScreen();
            continue;
        }

        if (in == PGU) {
            if (get_maxy() - 7 < 0)
                continue;
            if (netstartcounter - get_maxy() - 7 < 1)
                netstartcounter = 1;
            else
                netstartcounter -= get_maxy() - 7;
            updateScreen();
            continue;
        }
        //left(0, format("KEY %c    ", in));
    }
}

char range[128];
void updateScreen(void){
    long maxnets, net, cast, curr, max;
    int pct, y, column, h = 0, sw_searchresult;
    long count;
    char **show, *hlp;
    cursor_on(0);
    reverse(0);
    y = 0;

    sw_searchresult = strlen(foundres);

    switch (THESCREEN) {
    case 1:			//help
    case 2:			//vocabularium
    case 3:			//about
    case 4:			//command line help
        clear();
        show = helper;
        if (THESCREEN == 2)
            show = dictionary;
        if (THESCREEN == 3)
            show = about;
        if (THESCREEN == 4) {
            center(y, "COMMAND LINE ARGUMENTS");
            y++;
            drawhorline(y);
            y+=2;
            if(*err_stderr) {
                left(y, err_stderr);
                y+=2;
            }

            show = command_help;
        }
        while (NULL != (hlp = show[h])) {
            switch (hlp[0]) {
            case '-':
                drawhorline(y);
                break;
            case 'l':
                left(y, hlp + 1);
                break;
            case 'r':
                right(y, hlp + 1);
                break;
            case 'c':
                center(y, hlp + 1);
                break;
            default:
                left(y, hlp + (THESCREEN == 4 ? 0 : 1));
            }
            h++;
            y++;
        };
        break;

    default:			//subnet list
        //titels
        move(y, 0);
        clrtoeol();
        left(y, format(" SUBNET CALCULATOR version %d.%d", VERSION_HI, VERSION_LO));
        right(y, "[h]elp, [q]uit ");
        y++;
        drawhorline(y);

        y++;
        move(y, 0);
        clrtoeol();
        if (THETAB == 0) reverse(1);
        mvaddstr(y, 0, format(" cidr: %s", longtoip(ip)));
        if (THETAB == 0) reverse(0);

        mvaddstr(y, get_x(), " /");

        if (THETAB == 1) reverse(1);
        mvaddstr(y, get_x(), format(" %d ", bits_mask(mask)));
        if (THETAB == 1) reverse(0);

        if (THETAB == 2) reverse(1);
        x_mask = center(y, format("mask: %s", longtoip(mask)));
        if (THETAB == 2) reverse(0);

        if (THETAB == 3) reverse(1);
        x_wild = right(y, format("wild: %s ", longtoip(wild)));
        if (THETAB == 3) reverse(0);

        y++;
        move(y, 0);
        clrtoeol();

        if (THETAB == 4) reverse(1);
        left(y, format(" #nets: %u", getnets()));
        if (THETAB == 4) reverse(0);

        if (THETAB == 5) reverse(1);
        x_hosts = center(y, format("#hosts: %d ", gethosts()));
        if (THETAB == 5) reverse(0);

        right( y, format( "From: %s ",  longtoip( ipcalcnet(1) )));

        y++;
        move(y, 0);
        clrtoeol();

        left(y, format(" #bits: %d", bits_subnet()));
        center(y, format("IP Class %c [%s] ", getipclass(ip), iptype(ip)));
        right( y, format( "Till: %s ", longtoip( ipcalcnet(getnets())) ));
        y++;
        drawhorline(y);
        if (sw_searchresult) {
            reverse(1);
            center(y, foundres);
            reverse(0);
        } else
            center(y, format("%s", getipnetbits()));
        y++;
        move(y, 0);
        clrtoeol();

        //maak de lijst
        column = (int) (get_maxx() - 80) / 4;
        if (column < 0) column = 0;

        x_colpos[0] = 0;
        mvaddstr(y, x_colpos[0], columns[0]);

        x_colpos[1] = column + x_colpos[0] + 10;
        move(y, x_colpos[1] - 2);
        addch(CHAR_VLINE);
        mvaddstr(y, x_colpos[1], format("%s (magic: %d)", columns[1], magicnumber() ));

        x_colpos[2] = column + x_colpos[1] + 18;
        move(y, x_colpos[2] - 2);
        addch(CHAR_VLINE);
        mvaddstr(y, x_colpos[2], columns[2]);

        x_colpos[3] = column + x_colpos[2] + 18;
        move(y, x_colpos[3] - 2);
        addch(CHAR_VLINE);
        mvaddstr(y, x_colpos[3], columns[3]);


        y++;
        drawhorline(y);


        y++;
        count = netstartcounter;
        maxnets = getnets();
        h = 0;
        while (y <= get_maxy()) {

            net = ipcalcnet(count);
            cast = ipcalcbcast(count);

            move(y, 0);
            clrtoeol();
            if (sw_searchresult && h == 0) reverse(1);
            mvaddstr(y, x_colpos[0], format("%u", count));
            mvaddstr(y, x_colpos[1], longtoip(net));
            mvaddstr(y, x_colpos[2], longtoip(cast));
            memset(range, 0, 128);
            sprintf(range, "%s - ", longtoip(net + 1));
            sprintf(range + strlen(range), "%s", longtoip(cast - 1));
            mvaddstr(y, x_colpos[3], range);
            if (sw_searchresult && h == 0) {
                reverse(0);
                h = 1;
            }
            move(y, x_colpos[1] - 2);
            addch(CHAR_VLINE);
            move(y, x_colpos[2] - 2);
            addch(CHAR_VLINE);
            move(y, x_colpos[3] - 2);
            addch(CHAR_VLINE);
            y++;
            count++;
            if (count > maxnets) break;
        };

        /* clear the rest of the screen */
        h = 0;
        if (y < get_maxy()) {
            do {
                move(y + h, 0);
                clrtoeol();
                h++;
            } while ((h + y) <= get_maxy());
            drawverline(y, x_colpos[1] - 2);
            drawverline(y, x_colpos[2] - 2);
            drawverline(y, x_colpos[3] - 2);
        }


        curr = netstartcounter + y - 9;
        max = getnets();
        pct = PROCENT(curr, max);
        mvaddstr(6, x_colpos[0] + 1, format("(%d%%)", pct));

        if (THETAB == 0) move(2, 7);
        if (THETAB == 1) move(2, strlen(format("cidr:  %s / %d", longtoip(ip), bits_mask(mask))) - 1);
        if (THETAB == 2) move(2, x_mask + 6);
        if (THETAB == 3) move(2, x_wild + 6);
        if (THETAB == 4) move(3, 8);
        if (THETAB == 5) move(3, x_hosts + 8);
    }

    if (sw_searchresult)
        memset(foundres, 0, 128);

    if (strlen(ipbox_title))
        ipbox(NULL, 0);		//redraw the ip input box
    if (strlen(intbox_title))
        intbox(NULL, NULL, 0);	//redraw the integer input box
    if (strlen(mbox_title))
        messagebox(NULL, NULL);	//redraw the message box
    if (strlen(dlgbox_title))
        dlgbox(NULL);		//redraw the dialog box
    cursor_on(1);
}



void histogram(unsigned long curr, unsigned long max){
    float pct;
    char hist[64];

    if (sw_exit) return;

    pct = PROCENT(curr, max);
    memset(hist, 0, 64);
    sprintf(hist + strlen(hist), "  >> WRITING TO FILE: %.2f%%... <<  ", pct);
    cursor_on(0);
    reverse(1);
    center(5, hist);
    reverse(0);
    cursor_on(1);
    yield();
#ifdef __LINUX__
    refresh();
#endif

}
